#include <algorithm>
#include <map>
#include <vector>
#include <omp.h>
#include <sstream>

using namespace std;

#include "constants.h"
#include "Read.h"
#include "TrainData.h"

#include "distsampling.h"

#include "utils.h"
#include "MC2.h"
#include "YY_stopwatch.h"

#include "BmCalc.h"
#include "Logger.h"


// ----------------------------------------------------------------------------
// This was called 'trainData' in earlier versions of this program.
//
void BmCalc::Run( 
                MultiReadVector    & map_1,          // Unique Reads
                MultiReadVector    & mapMultiRead,   // MultiReads
                ErrorScoreMap      & a,              // Error score <map>
                std::vector<int>   & positionV,      // Absolute Positions of each of the UniqueReads (sorted)
                JobInfo            & jobInfo,         // Info necessary for processing this job (output filenames, etc)
				    std::string sProbOutput,
			    	 bool boStrandness
                )
{

	int n_loci= mapMultiRead[0]->Size();
   YY::Stopwatch stopWatch;  double dTotalTime = 0.0;  stopWatch.start();   //cout << __FUNCTION__ << ":  stopWatch started." << endl;
   double dMCMCTime = 0.0;        // Elapsed time spent in the mcmc function we call in our loop.
   double dLastReportTime = 0.0;  // Previous total time (accumulated from previous loops)

   Logger & lg = GetLogger();     // This MUST be a reference!
   std::ostringstream os;         // Used with the Logger.

   int M = _iSeqLen; 

   int iNumProb= -1;
   if(sProbOutput== "BOTH")
   {
	   iNumProb= 2;
   }
   else if (sProbOutput== "BAYES" || sProbOutput== "PROP")
   {
	   iNumProb= 1;
   }
   //cout << __FUNCTION__ << ": Thrd(" << jobInfo.iCalcId << ")  The size of mapMultiRead " << n_loci << " is: (" << mapMultiRead.size() << ") entries." << endl;
         
   // Write file header, now is removed to facilitate later file merging
   /*
   *(jobInfo.posCallRecStream)  << "ID"         << '\t'
                             << "Genome"     << '\t'
                             << "Position"   << '\t'
                             << "naive.call" << '\t'
                             << "b.call"
                             << std::endl;
							 */
  
   double ** hbetas = 0;
   ::Make2DArray(hbetas, _iTotalIterations, n_loci*_iSeqLen);   // Note: this is a 'new'ed array:  our function is responsible for 'delete'ing it.
                       
   bool ** Z_bools = 0;                        // We only need this for ONE MultiRead calculation.  We'll re-use it for each of the items in our MultiRead collection.
   ::Make2DArray(Z_bools, n_loci, _iSeqLen);   // Note: this is a 'new'ed array:  our function is responsible for 'delete'ing it.


   MCMCcalculator mcmc(_iTotalIterations, _iBurninIterations, _dPolymorphismRate);    // This is currently the only way we can set the Iterations, BurnInIters, and PolymorphismRate
   mcmc.SetNumLoci(n_loci);   
   mcmc.SetSeqLen(_iSeqLen);


   if( (jobInfo.iStartIndex < 0) || (jobInfo.iEndIndex > (int)mapMultiRead.size()) || (jobInfo.iStartIndex > jobInfo.iEndIndex) )
   {
     ostringstream os;
     os << __FUNCTION__ << ": Thrd(" << jobInfo.iCalcId << ")  ERROR: iStartIndex (" << jobInfo.iStartIndex << ") must be >= zero, and iEndIndex (" << jobInfo.iEndIndex << ") must be <= MultiRead vector size (" << mapMultiRead.size() << ")." << endl;
     throw std::invalid_argument( os.str() );
   }

 
   jobInfo.posCallRecStream->open(jobInfo.sCallRecFilename.c_str());       // These are the files to which we output our results.

#ifdef OUTPUT_POSTMEAN
   jobInfo.posPostMeanStream->open(jobInfo.sPostMeanFilename.c_str());     // TODO: error handling?
#endif

   // For each MultiRead....
   for( int i=jobInfo.iStartIndex; i < jobInfo.iEndIndex; ++i )  //i += n_loci )	//  These are indices within the mapMultiRead
   {           
      vector<int> t_v(n_loci);     // trcsize values 

   // [iLoci]  [k]  [iUnqRead]  MisMatchData
      vector< vector< vector<   MisMatchData  > > > vMisMatchIndicators;   // (was: 'tr')  vMisMatchIndicators[iLoci][k][iUnqRead][three0s]

      vector<string> vProfileStrs(n_loci);     // (was: xu) Error Profile strings, will be replaced with vMisMatches in Read structure

      // For each loci in this item in our multiRead mapMultiRead...  (look for UniqueReads with over-lapping SEQ_LEN Positions)
      for(int loci=0; loci<n_loci; loci++)
      {
	      vProfileStrs[loci] = mapMultiRead[i]->Reads(loci).sProfile;     // (was: xu)

         vector<vector< MisMatchData > > mmiPerLoci;     // 'mmi' means MisMatchIndicators   (previously called 't2')
         int upFlankingInPositionV = 0;
         int lowFlankingInPositionV = 0; 

         t_v[loci] = countMatchingUniqueReads_at_flanking(map_1, mapMultiRead[i]->Reads(loci), positionV, boStrandness, upFlankingInPositionV, lowFlankingInPositionV); 
         // Return number of UniqueReads that fall within this mulitRead and the flanking region (flanking region length could be 0) 
         // Up and Low identify the start and end index of positionV, so that later we only need to search within a much smaller range for each position in multiRead
                    
         //cout<<"searching between: "<<lowFlankingInPositionV <<" - "<< upFlankingInPositionV<< "\t"<< positionV[lowFlankingInPositionV]<<" - "<<positionV[upFlankingInPositionV]<< endl;
         //cout<< t_v[loci]<< " counts were found within above range."<<endl;

         for(int k=0; k<M; k++)		//  k = 0..34 -->  35 36 iterations
         {

            vector< MisMatchData > mmiAt_k;     //   vector of [iMisMatchPos,boMisMatch,sProfile] entries.)
            if (t_v[loci] > 0)  //search uniqueReads mapped to each position of the multiread only when there are such uniqueReads mapped to the flanking region
            {
                
                FindMatchingUniqueReads_at_k_2( k, 
                    map_1, 
                    mapMultiRead[i]->Reads(loci), 
                    positionV, 
                    upFlankingInPositionV, 
                    lowFlankingInPositionV,
                    mmiAt_k,    //the only output
                    boStrandness);         
            }           
           
            mmiPerLoci.push_back( mmiAt_k );  // Push a vector containing MisMatch Indicators for each MatchingUniqueRead at this 'k'.
                                              //  'mmi' means MisMatchIndicators      (formerly called 't2')
         }   // for(int k=0; k<M; k++)
         
         vMisMatchIndicators.push_back( mmiPerLoci );    // 'mmi' means MisMatchIndicators   (previously called 'tr')                 

      }  // for(int loci=0; loci<n_loci; loci++)
      //cout<< "before mcmc is fine!"<<endl;

      // We are now finished building  vMisMatchIndicators[iLoci][k][iUnqRead][three0s]
      //dTime = stopWatch.read();   cout << __FUNCTION__ << ":  Finished with finding MisMatchIndicators for a MultiRead: " << dTime << " secs, elapsed." << endl;
      /* --------------------------------------------------------- */

      //Apply posterior inference to trains[i]
      vector<int> mism(n_loci);    //int mism[n_loci];
      int min_naive;


      // Construct our bool array of indicators for where in the k's the mismatches occur:
      for(int iLoci=0; iLoci<n_loci; ++iLoci)
      {
         ::memset(Z_bools[iLoci], 0, _iSeqLen);  // Clear it out, set to false
         int iMMlen = iMISMATCH_LEN;             // Number of items in vMisMatches
         for(int m=0; m<iMMlen; ++m)
         {
            int iPos = mapMultiRead[i]->Reads(iLoci).vMisMatches[m].iPos;  // This iPos could be from 0 to _SeqLen.  (-1 means 'no mismatch')
            if(iPos >= 0) // If there is a mismatch at this position (k)
                Z_bools[iLoci][iPos] = true;
         }
         //cout << __FUNCTION__ << ": Z_bools: ";  for(int k=0; k<_iSeqLen; ++k) {cout << Z_bools[iLoci][k] << ", " ; }  cout << endl;
      }


      for (int j=0; j<n_loci; j++)
      {
         int sum=0;
         for(int k=0; k<M; k++)
            sum += Z_bools[j][k];  // Z[i/n_loci][j][k];  // TODO:  'bool to int' conversion is OK?
         mism[j]=sum;
      }

      // Initialize our hbetas array to some near-zero values.
      for (int j = 0 ; j < _iTotalIterations ; j++)
      {  
         for (int k = 0 ; k < n_loci * M ; k++)
         {  
            hbetas[j][k] = 0.000001;   // This is a 'magic number' that is NON-ZERO, and identifies this as having not been calculated.
         }
      }


      vector<double> prob(n_loci, 0.0);    //prob is a vector of n_loci entries

      // The MCMC calculation: (returns results in 'prob' and a portion of 'hbetas')
      if(sProbOutput== "BOTH" || sProbOutput== "BAYES") //selectively perform mcmc calculation. proportional probability can always be calculated because it takes little time
      {
         dMCMCTime += mcmc.Run( vMisMatchIndicators, Z_bools, a, hbetas, prob, vProfileStrs, t_v ); // Apply mcmc here		
      }


      min_naive = min_value(mism);  
      vector<int> min_id;
	   int trc_size_sum=0;   // total number of unique reads mapped to top hits
      for(int j=0;j<n_loci;j++)
      {
         if (mism[j] == min_naive)
		   {
            min_id.push_back(j);
			   trc_size_sum += t_v[j];
		   }		 
      }

      vector<double> naive_call(n_loci,0.0);   // the naive probability    (initialized to zero)
	   vector<double> prop_call(n_loci,0.0);    // the proportional probability

       for(int j=0; j<(int)min_id.size(); j++)
      {
         int idx = min_id[j];
		   naive_call[idx] = (double)1/(min_id.size());
		   //calculate proportional probability based on # of unique reads mapped to top hits of unique reads
		   if(trc_size_sum > 0)
		   {
			   prop_call[idx]= (double)t_v[idx] / (double)trc_size_sum;
		   }
		   else
		   {
			   prop_call[idx]= naive_call[idx];
		   }
      }

      
      for(int j=0; j<n_loci; j++)
      {
		  *(jobInfo.posCallRecStream)    //<<mapMultiRead[i]->Reads(j).iLineNumberInSAMFile<< "\t"
										 << mapMultiRead[i]->Reads(j).sID         << "\t"
			                      << mapMultiRead[i]->Reads(j).sChrom      << "\t"        
			                      << mapMultiRead[i]->Reads(j).iPosition   << "\t"
                               << naive_call[j] << "\t"  
									    << prop_call[j]<< "\t" << prob[j]                            
                               << std::endl;


         // These are our results!   This is the entire point of this program.
         if (sProbOutput== "BOTH")
         {
	         mapMultiRead[i]->Reads(j).dProbs[0] = prop_call[j];    // Proportional probability  // NOTE: 0 and 1 array positions
	         mapMultiRead[i]->Reads(j).dProbs[1] = prob[j];         // Bayes probability
         }
         else if(sProbOutput== "BAYES")
         {
	         mapMultiRead[i]->Reads(j).dProbs[0] = prob[j];
         }
         else  // PROPORTIONAL
         {
	         mapMultiRead[i]->Reads(j).dProbs[0] = prop_call[j];
         }
      }
     

      vector<double> bmean( M*n_loci, 0.0 );
      for(int j=0; j<n_loci*M; j++)
      {
         bmean[j]=0.0;
         for(int k=0; k<( _iTotalIterations - _iBurninIterations+1); k++)
         {
            bmean[j] = bmean[j] + hbetas[_iBurninIterations+k-1][j];
         }
         bmean[j] = bmean[j]/(_iTotalIterations - _iBurninIterations+1);
      }

#ifdef OUTPUT_POSTMEAN      
      *(jobInfo.posPostMeanStream) << mapMultiRead[i]->Reads(0).sID << '\t';
      for(int j=0; j<n_loci*M; j++)
      {
         *(jobInfo.posPostMeanStream) << bmean[j] << '\t';   
      }
      *(jobInfo.posPostMeanStream) << endl;
#endif      


      // Output some diagnostic information:  (to the console ONLY!)
      if( (i % 5000) < 1 )     // Output a line for every 5000 MultiRead containers we process.
      {
         dTotalTime = stopWatch.read();
         double dDelta = dTotalTime - dLastReportTime;        // Elapsed time since we last output a report (here).
		 //cout<< __FUNCTION__ << ": On actual thread: "<< th_id << ": " << endl; // output the current thread number
         cout << __FUNCTION__ << ": Thrd(" << jobInfo.iCalcId << "),  MultiRead[" << i << "]\t Elapsed time in this function: " << dTotalTime << " secs" 
                              << "   In MCMC: " << dMCMCTime << " secs" << "   SinceLastReport: " << dDelta  << " secs" << endl;
         dLastReportTime = dTotalTime;   // time now.
      }


   } // for( int i=jobInfo.iStartIndex; i < jobInfo.iEndIndex; ++i )

   ::Delete2DArray(hbetas, _iTotalIterations);
   ::Delete2DArray(Z_bools, n_loci);  


   jobInfo.posCallRecStream->close();    // These are the files to which we output our results.
#ifdef OUTPUT_POSTMEAN
   jobInfo.posPostMeanStream->close();   // These are the files to which we output our results.
#endif
   //jobInfo.posTrcSizeStream->close();

   dTotalTime = stopWatch.read();   
   lg.Write( ((os << __FUNCTION__ << ": Thrd(" << jobInfo.iCalcId << ")   Total elapsed time in this function: " << dTotalTime << " secs   (" << dTotalTime/60.0 << " mins)" << endl), os ));
}

// ----------------------------------------------------------------------------
// Function return value: the number of UniqueReads that fall within the 
//   multiRead + flanking areas.
int BmCalc::countMatchingUniqueReads_at_flanking(
											const MultiReadVector & map_1,				   // input: Unique Reads, now in a MultiRead vector
											const Read				 & multiRead,           // input: MultiRead we're working with
											vector<int>				 & positionV,           // input: Positions of each of the UniqueReads (sorted)
											bool boStrandness,                           // default value is false: no need to consider strandness
                                 int & upFlankingInPositionV,                 // output: the up bound from previous search for uniqueRds mapped to flanking region
                                 int & lowFlankingInPositionV                 // output: the low bound from previous search for uniqueRds mapped to flanking region
										)
{
    int iUniqueRds_count = 0;
	int b  = multiRead.iPosition;                 // This is the Position of this read on the genome: like 17190096
	int up  = b +  _iSeqLen -1 +  _iSeqFlankingValue;        // Widen the area we search to include the FlankingValue on each side of the the multiRead
	int low = b - (_iSeqLen +     _iSeqFlankingValue -1);    // _iSeqFlankingValue could be 0 
	vector<int>::iterator it_low = upper_bound( positionV.begin(), positionV.end(), low-1);   // upper_bound function returns the first number in positionV that is strictly greater than low-1, or greater than/equal "low"
	vector<int>::iterator it_up  = upper_bound( it_low           , positionV.end(), up   )-1; // upper_bound function returns the first number in positionV that is strictly greater than up, -1 to get the last number that is less than/equal "up"
	lowFlankingInPositionV = (int) std::distance(positionV.begin(), it_low);
	upFlankingInPositionV  = (int) std::distance(positionV.begin(), it_up );
	bool boSameStrand= true;

	for(int i= lowFlankingInPositionV; i<= upFlankingInPositionV; i++)
	{
		string sChrom1 = map_1[i]->Reads(0).sChrom;
		string sChrom2 = multiRead.sChrom;
		if (boStrandness) // must check whether the unique reads are on the same strand as multiRead
		{
			char cStrand1= map_1[i]->Reads(0).cStrand;
			char cStrand2= multiRead.cStrand;
			boSameStrand= (cStrand1== cStrand2);
		}
		
		if(sChrom1 == sChrom2 && boSameStrand)
		{
            iUniqueRds_count++;		
      }
	}
	return iUniqueRds_count;
}

// Another version of FindMatchingUniqueReads_at_k that uses a more efficient searching algorithm.
// Scenarios to test:
//  - Make SURE the algorithm works with proper input (proper means 'up' and 'low' are within the range of values in positionV)
//  - Make sure the algorithm works with 'up' and 'low' values that are actual numeric position entries in positionV
//  - What if 'up' and/or 'low' are NOT contained within the range of values in positionV  ?
//  - What if positionV is empty ?
//  - What if 'k' is negative or larger than _SeqLen  (this would be due to an error somewhere else in the program!) ?
//  - This has nothing to do with flanking

size_t BmCalc::FindMatchingUniqueReads_at_k_2(
											const int k,                                 // input, the relative position, from 0- _iSeqLen
											const MultiReadVector & map_1,					// input: Unique Reads, now in a MultiRead vector
											const Read				 & multiRead,           // input: MultiRead we're working with
											vector<int>				 & positionV,           // input: Positions of each of the UniqueReads (sorted)
                                 const int upFlankingInPositionV,             // input: the up bound from previous search for uniqueRds mapped to flanking region
                                 const int lowFlankingInPositionV,            // input: the low bound from previous search for uniqueRds mapped to flanking region
                                 std::vector< MisMatchData >	 & mmiAt_k,    // output
											bool boStrandness  //default value is false: no need to consider strandness
										)
{
	int b  = multiRead.iPosition;      // This is the Position of this read on the genome: like 17190096
	int up  = b+k;         
	int low = b+k - (_iSeqLen -1);    
	vector<int>::iterator it_low = upper_bound( positionV.begin()+ lowFlankingInPositionV, positionV.begin()+ upFlankingInPositionV, low-1);   // upper_bound function returns the first number in positionV that is strictly greater than low-1, or greater than/equal "low"
	vector<int>::iterator it_up  = upper_bound( it_low           , positionV.begin()+ upFlankingInPositionV, up   )-1; // upper_bound function returns the first number in positionV that is strictly greater than up, -1 to get the last number that is less than/equal "up"
	int idx_low = (int) std::distance(positionV.begin(), it_low);
	int idx_up  = (int) std::distance(positionV.begin(), it_up );
	bool boSameStrand= true;

	for(int i= idx_low; i<= idx_up; i++)
	{
		string sChrom1 = map_1[i]->Reads(0).sChrom;
		string sChrom2 = multiRead.sChrom;
		if (boStrandness) // must check whether the unique reads are on the same strand as multiRead
		{
			char cStrand1= map_1[i]->Reads(0).cStrand;
			char cStrand2= multiRead.cStrand;
			boSameStrand = (cStrand1 == cStrand2);
		}
		
		if( (sChrom1 == sChrom2) && (boSameStrand) )
		{
            FindMisMatches_at_k(k, multiRead, map_1[i]->Reads(0), mmiAt_k); //add one MisMatchData object to mmiAt_k each time
		}
	}

	return mmiAt_k.size();
}

// ----------------------------------------------------------------------------
// This function returns a vector of MisMatchData objects (mmiAt_k), with each object 
//  corresponding to an item in the vMatchingUniqueRdsAt_k vector.
//  The .boMisMatch field in each MisMatchData object is set according to
//  whether there is a mismatch in the UniqueRead at position 'k'.  (k goes from 0 to iSeqLen-1).
// We are still working with data at just a single 'k', within the multiRead.
// notice the argument changes from a vector of unique read to one single unique read so that it can be used in FindMatchingUniqueReads_at_k_2
size_t BmCalc::FindMisMatches_at_k( 
                                 const int k,                                       // input: index within the multiRead (0 to 35)
                                 const Read                     &  multiRead,       // input: MultiRead we're working with
                                 Read                           &  uniqueRead,      //input: uniqueRead that mapped to the kth position of MultiRead  
                                 std::vector< MisMatchData >	 & mmiAt_k           // output
                               )
{
	    
                              // iMisMatchPos, boMisMatch, sProfile
   MisMatchData misMatchData = { 0           , false     , ' '  };   // Used just for initialization of items we push_back onto mmiAt_k

   mmiAt_k.push_back(misMatchData);      // Initialize it, then alter it below.
                                              
   //MisMatchData * pMis = &(mmiAt_k[t]);   // Debug

   int iMatchingUniqueRd_AbsPosition = uniqueRead.iPosition ;   // Absolute position of the start of the Read within the genome.

   int iAbsPosAt_k = multiRead.iPosition + k; //absolute position of our interest within the genome
   int iMatchingUniqueRd_PosWithinRead =  iAbsPosAt_k - iMatchingUniqueRd_AbsPosition ; ////with flanking option, now it is necessary to check whether the iPosWithinRead is valid. For example, the unique reads overlap with the flanking regons only should not be included in mmiAt_k.

   size_t t= mmiAt_k.size() -1; // t is the index of newly pushed item in mmiAt_k 
   mmiAt_k[t].iPosWithinRead = iMatchingUniqueRd_PosWithinRead;    // Relative pos from the beginning of the Read
	  	  
   int iMMIndex= uniqueRead.FindMisMatchIndexAtAbsolutePos(iAbsPosAt_k);
   if (iMMIndex!=-1 && iMMIndex< iMISMATCH_LEN )
   {      
	   mmiAt_k[t].boMisMatch = true;    // MisMatch 
	   mmiAt_k[t].cMis   = uniqueRead.vMisMatches[iMMIndex].chError;   // Example: "DLDPTTHHHVGPPGTPVTPVVVVVVVVVVVVVVTVV"		
   }
   else
   {
	   mmiAt_k[t].boMisMatch = false;     // No mismatch
   }

   return mmiAt_k.size();
}

