
#ifndef _BMCALC_
#define _BMCALC_

#include <algorithm>  // for std::lower_bound
#include <vector>
#include "Read.h"
#include "utils.h"

// ----------------------------------------------------------------------------
// Do a binary search between [first,last), returning the iterator pointing
//   to the 'found' Value.  If the Value is not found, the returned iterator
//   will be pointing to 'last'.
template <class ForwardIter, class Type>
ForwardIter BsearchAndReturn(  ForwardIter first,  ForwardIter last, const Type& crValue)
{
   first = std::lower_bound(first, last, crValue);
   return ((first == last || crValue < *first) ? last : first);
}


// ----------------------------------------------------------------------------
// Bayesian Mapping Calculator

class BmCalc
{

private:
   int    _iSeqLen;
   int    _iSeqFlankingValue;
   double _dPolymorphismRate;  // For the MCMC calculations
   int    _iTotalIterations;   // For the MCMC calculations
   int    _iBurninIterations;  // For the MCMC calculations


   // find the total number of unique reads mapped to the (multiread + flanking region)
   int countMatchingUniqueReads_at_flanking(
											const MultiReadVector & map_1,						// input: Unique Reads, now in a MultiRead vector
											const Read				 & multiRead,              // input: MultiRead we're working with
											vector<int>				 & positionV,              // input: Positions of each of the UniqueReads (sorted)
											bool boStrandness,  //default value is false: no need to consider strandness
                                            int & upFlankingInPostionV,                 //output: the up bound from previous search for uniqueRds mapped to flanking region
                                            int & lowFlankingInPositionV               //output: the low bound from previous search for uniqueRds mapped to flanking region
											
											//int iflanking
										);

   // Another version of FindMatchingUniqueReads_at_k
   // This version will use the STL algorithm upper_bound and lower_bound
   // The updated version of FindMatchingUniqueReads_at_k_2 also populates mmiAt_k, the vector of MisMatchData what stores the mismatch information at  
   // this corresponding position of the unique read
   size_t FindMatchingUniqueReads_at_k_2(
											const int k,										// input
											const MultiReadVector	& map1,						// input: Unique Reads, now in a MultiRead vector
											const Read				& multiRead,				// input: MultiRead we're working with
										    vector<int>				& positionV,				// input: Positions of each of the UniqueReads (sorted)
                                            const int upFlankingInPostionV,                 //input: the up bound from previous search for uniqueRds mapped to flanking region
                                            const int lowFlankingInPositionV,               //input: the low bound from previous search for uniqueRds mapped to flanking region
											// map<string, bool>		& mapUniqueTrc_IDs_flanking,			// output: extended to flanking region
											// MultiReadVector			& vMatchingUniqueRdsAt_k,	// output: flanking = 0, contain uniqueRds that overlap with multiRead
                                            std::vector< MisMatchData > & mmiAt_k,          //output
											bool boStrandness  //default value is false: no need to consider strandness
											//int iflanking   //could be set to number not equal _iSeqFlankingValue
										);

   // FindMisMatches_at_k will be used in FindMatchingUniqueReads_at_k_2
   size_t FindMisMatches_at_k( 
                           const int k,                                             // input: index within the multiRead (0 to 35)
                           const Read                    &  multiRead,               // input: MultiRead we're working with
                           //const MultiReadVector        & vMatchingUniqueRdsAt_k,   // input
                           Read                          & uniquRead, //input: uniqueRead that mapped to the kth position of MultiRead
                           std::vector< MisMatchData >   & mmiAt_k              // output
                           );
      

public:
   // ctors:
   BmCalc( )
      : _iSeqLen(35), _iSeqFlankingValue(0), _dPolymorphismRate(0.1), _iTotalIterations(100), _iBurninIterations(20)
   {  };

   BmCalc( int iSeqLen, int iSeqFlankingValue, double dPolymorphismRate, int iTotalIterations, int iBurninIterations)
      : _iSeqLen(iSeqLen), _iSeqFlankingValue(iSeqFlankingValue), _dPolymorphismRate(dPolymorphismRate), _iTotalIterations(iTotalIterations), _iBurninIterations(iBurninIterations)
   {  };



void Run( 
                MultiReadVector & map_1,          // Unique Reads
                MultiReadVector & mapMultiRead,   // MultiReads
                ErrorScoreMap        & a,             // Error scores 
                std::vector<int>     & positionV,      // Positions (absolute, within the gene sequence) of each of the UniqueReads (sorted)
                JobInfo              & jobInfo,
   				 std::string sProbOutput,
				    bool boStrandness
               );


};


#endif
