
#ifndef _CONFIGFILEREADER_H_
#define _CONFIGFILEREADER_H_
                   
#include <string>
#include <map>
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

// See the large comment section at the bottom of this file for additional
//  documentation about what this class does.
//

// ----------------------------------------------------------------------------
// This class reads a plain input (configuration) text file consisting of Name/Value pairs. 
// It places the Name/Value pairs in an stl::map so that the Values can
// be easily retrieved by users of this class.
// The input file has this format (example):
//   # ---
//   # Comments follow the "#" character
//   #   
//   PROBABILITY = 0.01
//   INPUT_DIR   = ./in_dir
//   OUTPUT_DIR  = ./out_dir
//   POLYMORPHISM_RATE = 0.001
//   STEP_SIZE    = 4
//   # ---


class ConfigFileReader
{

private:   // member data:   
           // (Note leading underscore ("_")  on names indicating 'member variables')   

   typedef  std::map<std::string, std::string>  NVMAP_TYPE;
   std::string  _sFilename;
   NVMAP_TYPE   _mapNameValuePairs;

   char  _chCommentChar;     // In the input file (chars following this are ignored)
   char  _chNewLine    ;
   char  _chDelim      ;    // Between the Name and Value in the input file


public:    // functions

   ConfigFileReader() 
   {
      DefaultInit();
   };

   ConfigFileReader( std::string const sFilename ) 
   { 
      DefaultInit();
      SetFilename(sFilename);  
   };

   ConfigFileReader( ConfigFileReader const & reader )     // Copy constructor
   {  
      _mapNameValuePairs.clear();
      _mapNameValuePairs = reader._mapNameValuePairs;  //  TODO: verify that this works OK!
      _sFilename         = reader._sFilename;

      _chCommentChar = reader._chCommentChar;    // In the input file (chars following this are ignored)
      _chNewLine     = reader._chNewLine;
      _chDelim       = reader._chDelim;          // Between the Name and Value in the input file

   };

   void DefaultInit()
   {
      _chCommentChar = '#';     // In the input file (chars following this are ignored)
      _chNewLine     = '\n';
      _chDelim       = '=' ;    // Between the Name and Value in the input file
   };

   virtual ~ConfigFileReader() {};


   // --------------------------------------------------------------
   // Return value: true if we found sName in our map, false otherwise.
   template<typename T>
   bool GetValue(std::string const sName, //  input
                 T & value                //  output
                )
   {
      using namespace std;
      bool boFoundName = true;

      NVMAP_TYPE::iterator it =  _mapNameValuePairs.find(sName);
      if( it != _mapNameValuePairs.end() )    // If we found it...
      {
         string sValue = it->second;     
         std::istringstream  is(sValue);
         is >> value;        
      }
      else
         boFoundName = false;

     return boFoundName;
   };

   // --------------------------------------------------------------
   void SetFilename(std::string const sInputFilename)
   { _sFilename = sInputFilename;  }


   // --------------------------------------------------------------
   string GetFilename()
   { return _sFilename ;  }


   // --------------------------------------------------------------
   size_t GetValueCount()
   {
      return _mapNameValuePairs.size();
   };


   // --------------------------------------------------------------
   virtual string ToString()
   {
      ostringstream  osRtn;
      NVMAP_TYPE::iterator it    =  _mapNameValuePairs.begin();
      NVMAP_TYPE::iterator itEnd =  _mapNameValuePairs.end();
      for(     ; it != itEnd; ++it )
      {
         //                   Name       =          Value
         osRtn << "  " << it->first << " = " << it->second  << endl;    
      }

      return osRtn.str();
   };



   // ---------------------------------------------------------------
   // Read name/value pairs from the file, put them into our map.
   // Returns the number of Name/Value pairs read from the input file.
   // The input file has this format (example):
   // # ---
	// # Comments follow the "#" character
	// #   
	// PROBABILITY = 0.01
	// INPUT_DIR   = ./in_dir
	// OUTPUT_DIR  = ./out_dir
	// POLYMORPHISM_RATE = 0.001
	// STEP_SIZE    = 4
	// # ---
   //
   virtual size_t ReadInputFile()
   {
      using namespace std;
      std::string sLine;
      std::vector<std::string> vLines;
      
      std::ifstream fileStream( _sFilename.c_str() );  // Open the file

      std::getline(fileStream, sLine, _chNewLine);   // Read a line from the file.
      while( !TestStreamFlags(fileStream) )         // Non-zero return value means the stream is in a bad state (or EOF)
      {   
         //std::cout << sLine << std::endl;

         TrimSpaces(sLine);
         size_t iPos = sLine.find( _chCommentChar );   // Look for a Comment char. 
         if(iPos != std::string::npos )
         {
            sLine.erase(iPos);            // Remove chars following the Comment char.
         }

         if(sLine.length() > 0)           // If there is still anything left in the string...
         {
            vLines.push_back(sLine);      // Save it.
         }
         std::getline(fileStream, sLine, _chNewLine);  // Read another line from the file.
      }
      fileStream.close();                 // We are finished reading the file.

      //cout << "----------" << endl;

      if( vLines.size() > 0 )          // Don't clear the existing map until we successfully read something from the file.
      {                                       
         _mapNameValuePairs.clear();   // Start with an empty map, in case we already read it previously.
      }

      //cout << __FUNCTION__ <<  endl;       // Diagnostic info 

      // Separate each of the strings, then put the Name/Value pairs in our map:
      for( vector<string>::iterator it = vLines.begin(); it != vLines.end(); ++it)
      {
         string sStr = *it;
         size_t iPos = sStr.find(_chDelim);   // This is the delimiter between the Name and Value (usually '=')
         if(iPos != string::npos)             // If we found the delimiter ...
         {
            string sKey   = sStr.substr(0,iPos);     TrimSpaces(sKey);
            string sValue = sStr.substr(iPos+1);     TrimSpaces(sValue);
            //cout << "  " << sKey << " = " << sValue  << endl;       // Diagnostic info 
            _mapNameValuePairs[sKey] = sValue;                      // Add the Name/Value pair to our collection
         }
      }

      return _mapNameValuePairs.size();
   };

 
private:   // functions

   // ----------------------------------------------------------------------------
   // Returns zero if the input stream is in a 'good' state,
   //   non-zero otherwise.
   int TestStreamFlags( std::ios & x ) 
   {
       int iRtn = 0, iState = x.rdstate( );
       iRtn = iState & std::ios::badbit ;
       if( iRtn )
           return iRtn;
            
       iRtn = iState & std::ios::failbit;
       if( iRtn )
           return iRtn;

       iRtn = iState & std::ios::eofbit;
       return iRtn;
   }

   // ------------------------------------------------------------------------
   // Trim spaces and tabs from the beginning and end of the string.
   void TrimSpaces( std::string & str)  
   {  
      // Trim Both leading and trailing spaces  
      size_t startpos = str.find_first_not_of(" \t");   // Find the first character position after excluding leading blank spaces  
      size_t endpos   = str.find_last_not_of (" \t");   // Find the first character position from reverse af  

      // If all spaces or empty return an empty string  
      if(( std::string::npos == startpos ) || ( std::string::npos == endpos))  
      {  
         str = "";  
      }  
      else 
         str = str.substr( startpos, endpos-startpos+1 );  

      /*  
      // Code for  Trim Leading Spaces only  
      size_t startpos = str.find_first_not_of( \t); // Find the first character position after excluding leading blank spaces  
      if( string::npos != startpos )  
         str = str.substr( startpos );  
      */ 

      /*  
      // Code for Trim trailing Spaces only  
      size_t endpos = str.find_last_not_of( \t); // Find the first character position from reverse af  
      if( string::npos != endpos )  
         str = str.substr( 0, endpos+1 );  
      */ 

   }  

   

};   // end Class ConfigFileReader

/*  ---------------------------------------------------------------------------

                        Software Component Specification
                                 08Sept2010

						
Component name: "ConfigFileReader"
Type: Class 
Language: C++ (ANSI, native)

Purpose:
   The ConfigFileReader class reads Name/Value pairs from a text file.  The 
   Values    are made available to the software developer by accessing each
   Name as a string (such as "PROBABILITY"). 

Usage (within the C++ program):
  line 1)  ConfigFileReader myReader;
       2)  myReader.SetFilename("InputInfo.txt");
       3)  myReader.ReadInputFile();
       4)  double dProb;
       5)  myReader.GetValue("PROBABILITY", dProb);
       6)  string sOutputDir;
       7)  myReader.GetValue("OUTPUT_DIR", sOutputDir);
	   8)  int iNumValues = myReader.GetValueCount();

Config File description:
  The input file that this class reads has this format (example):
    # ---
	# Comments follow the "#" character
	#   
	PROBABILITY = 0.01
	INPUT_DIR   = ./in_dir
	OUTPUT_DIR  = ./out_dir
	POLYMORPHISM_RATE = 0.001
	STEP_SIZE    = 4
	# ---

  Note that this is NOT an XML file.
  
  The separator between the Name and Value is the "equal" sign (=).
  The Name is case sensitive. It is suggested that it always be specified in
  UPPERCASE, to avoid potiental errors when calling the 
  ConfigFileReader::GetValue(...) function.

  There is not a limit on the number of lines or Name/Value pairs in the text
  file.


The ConfigFileReader::GetValue(...) function is overloaded so that it can 
fetch values of various types (long,double,string,etc), based on the type of 
the second argument. Note that it is possible that an error could occur 
when converting the Value read from the config file to the desired type.  
Under these circumstances, a C++ exception will be thrown by the 
ConfigFileReader::GetValue function.  Callers should be prepared 
for this via the use of 'try/catch' blocks.


The ConfigFileReader::ReadInputFile(...) function will throw a C++ exception 
if it cannot open and read the contents of the input file.  Callers should 
also be prepared for this.
*/





#endif  

