#ifdef _WIN32
#include <windows.h>     // for Timing functions not available on Linux
#else
#include <sys/time.h>    // on non-WIN32 systems
#endif

#include <omp.h>
#include <ctime>
#include "Logger.h"

using namespace std;

Logger::Logger(void)
       : _console( std::cout ), _sFilename("Logger.out"), _boFileIsOpen(false)
{
}

Logger::Logger( char const * sFilename )
       : _console( std::cout ), _sFilename(sFilename), _boFileIsOpen(false)
{
}


Logger::~Logger(void)
{
   if(_boFileIsOpen)
      Close();
}

string Logger::Filename()
{
   return _sFilename;
}


// ----------------------------------------------------------------------------
// Write the contents of the passed-in stream to the appropriate devices.
bool Logger::Write( std::ostringstream & os,  bool boClearStrm /*= true*/ )
{
   bool boOK = Write(os.str());
   if(boOK)
   {
      if(boClearStrm)
      {
         ClearBuf(os);
      }
   }
   return boOK;
}


// ----------------------------------------------------------------------------
// Write the contents of the passed-in string to the appropriate devices.
// Also note:  we are writing error strings to the *console* (this might not be what you want).
bool Logger::Write( string const & sOutput )
{
   bool boOK = true;
   #pragma omp critical  // Only let one Logger at a time write to the file
   {                     // We are trying to serialize the writes to the output file.

      _console << sOutput;
      
      try
      {
         int th_id = omp_get_thread_num();
         bool boOK = Open();
         if(boOK)
         {
            _ofile  << "Thr:" << th_id << " " << TimeStamp() << "  " << sOutput ;  // Note: no std::endl, caller should supply that if he needs it.
            boOK = !_ofile.fail();
            _ofile.flush();
            Close();
         }
         else
         {
            boOK = false;
            std::string sErr = __FUNCTION__ ;  sErr += ": ERROR, The output file (";   sErr += _sFilename + ") could not be opened for writing. "  ;
            _console << std::endl << sErr << std::endl;    // TODO: better error-handling      
         }
      }
      catch( runtime_error & er)
      {
         std::string sErr = __FUNCTION__ ;  sErr += ": ERROR, We caught an exception while writing to the output file (";   sErr += _sFilename + "). "  ;
         _console << std::endl << sErr << std::endl << er.what() << std::endl;    // TODO: better error-handling     
         Close();
         boOK = false;
      } 
   }
   return boOK;
}


// ----------------------------------------------------------------------------
void Logger::ClearBuf(std::ostringstream & os) 
{ 
   os.clear(); os.str("");  // Need .clear() to clear the io flags too.
}


// ----------------------------------------------------------------------------
bool Logger::Open()         
{  
   _ofile.open(_sFilename.c_str(), std::ios_base::app);     // open for Appending text
   _boFileIsOpen = !(_ofile.fail()); 
   return _boFileIsOpen;
}  

// ----------------------------------------------------------------------------
void Logger::Close()        
{  
   _ofile.close();  
   _boFileIsOpen = false; 
}

// ----------------------------------------------------------------------------
// Create a new file having sFilename.
bool Logger::CreateOutFile(std::string const & sFilename) 
{  
   _sFilename = sFilename; 
    return CreateOutFile();
}

// ----------------------------------------------------------------------------
// Create a new file having _sFilename.
bool Logger::CreateOutFile() 
{   
   bool boItWorked = false;
   _ofile.open(_sFilename.c_str(), std::ios_base::out);   // Note: ios_base::out
   if( _boFileIsOpen =  !(_ofile.fail()) )
   {
      _ofile << "Log created: " << TimeStamp() << std::endl << std::endl;
      boItWorked =  !(_ofile.fail());
   }
   Close();
   return boItWorked;
};

// ----------------------------------------------------------------------------
std::string Logger::TimeStamp()
{
   std::string sTimeStamp;
#ifdef _WIN32
   SYSTEMTIME now;   
   GetLocalTime(&now); 
   std::ostringstream os;
   os << now.wYear << "-" << now.wMonth << "-" << now.wDay<< "_" 
      << now.wHour<< "h_"<< now.wMinute << "m_" << now.wSecond << "s_" << now.wMilliseconds << "ms";
   sTimeStamp = os.str();
#else
   // On a non-Windows platform, use the lame 'localtime' function
	struct tm *ti;
	time_t epochtime;
	char str[42];
	time(&epochtime);            // Get the number of seconds since 1970
	ti = localtime(&epochtime);  // and convert to a date/time structure

	sprintf(str,
			"[%04d%02d%02d__%02d_%02d_%02d]",
			ti->tm_year + 1900,
			ti->tm_mon + 1,
			ti->tm_mday,
			ti->tm_hour,
			ti->tm_min,
			ti->tm_sec);      // Sorry, we can only produce "second" resolution.
	sTimeStamp = str;
#endif

	return sTimeStamp;
}



