#ifndef _LOGGER_H
#define _LOGGER_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept> 

// ----------------------------------------------------------------------------
// This is a class used for logging a string to the console (cout) and
//   to a file.
// It uses OpenMP to serialize the output to the devices.
//
// Example usage:
/*
   #include "Logger.h"

   Logger lgmaster;
   bool boOK = lgmaster.CreateOutFile();  // Uses the default "Logger.out" filename.
   if(!boOK)
   {
      cout << __FUNCTION__ << ": ERROR, failed to Create the output file (" << lgmaster.Filename() << ")." << endl;
   }

   int j;
   Logger lg;    // Uses the default "Logger.out" filename.

   #pragma omp parallel for private (j)
   for(j=0; j<4; ++j)
   {                                   
      std::ostringstream os;    // Note: each 'os' should be on a separate thread
            
      for(int i=0; i<100; ++i)
      {        
         boOK = lg.Write( ((os << "This is a test of lg.Write(os), " << j << "," << i << endl), os) ); 
         
         os << "Writing another line...." << endl;
         lg.Write(os);      // Note: this will .clear the 'os'.
         lg.Write("You can write like this also.");
      }     
   }
*/


// ----------------------------------------------------------------------------
class Logger
{

private:
   std::ostream  & _console;
   std::ofstream   _ofile;
   std::string     _sFilename;  // for _ofile stream
   bool            _boFileIsOpen;

   Logger (Logger const & rhs) : _console(std::cout)  // Private: don't allow this
   { rhs; };                            
   Logger & operator = ( Logger const & rhs )         // Private: don't allow this
   { return *this; };                           

public:
   Logger(void);
   Logger( char const * sFilename );
   virtual ~Logger(void);
    
 //std::ostringstream  buf;     // This *public* so users can do this:  myLogger.buf << "Sending stuff to the logger." ;

   std::string Filename();

   bool Write( std::ostringstream & os,  bool boClearStrm = true );
   bool Write( std::string const & sOutput);

   void ClearBuf( std::ostringstream & os );
   bool Open() ;
   void Close();
   bool CreateOutFile( std::string const & sFilename );
   bool CreateOutFile(); 
   std::string TimeStamp();

} ;


#endif
