//
// MC2.CPP
//

// Turn off warnings about unsafe code in Microsoft libraries:
#define _SCL_SECURE_NO_WARNINGS

#include <cassert>
#include <cmath>
#include <string>
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <stdexcept>


using namespace std;
#include "utils.h"
#include "MC2.h"
#include "distsampling.h"
#include "YY_stopwatch.h"



// ----------------------------------------------------------------------------
// Create a 2D array having the specified rows and columns.
// Our destructor will delete the Array.
void MCMCcalculator::CreateCurrDraw(int iNRows, int iNCols)
{
   _iNumCurrDrawRows = iNRows;
   _iNumCurrDrawCols = iNCols;
   _curr_draw = 0;
   ::Make2DArray( _curr_draw, iNRows, iNCols);   // Create the array
   return;

}

MCMCcalculator::~MCMCcalculator()
{
   if(_curr_draw)
   {
      ::Delete2DArray(_curr_draw, _iNumCurrDrawRows );
   }

}


// ----------------------------------------------------------------------------
// The function return value is the elapsed time (in seconds) spent in this function.
//
double MCMCcalculator::Run(
           vector< vector< vector< MisMatchData > > > const & train,
           bool ** Z,
           ErrorScoreMap &a,         
           double ** hbetas,               // A portion of this is calculated and returned.  The dimensions of this are in _iSeqLen and _iNum_loci         #endif
           vector<double> & prob,          // This is calculated and returned.
           vector<string> const & xu,      // These are the Profile strings
           vector<int> const & t_v         // Number of UniqueReads overlapping each MulitRead (t_v.size() == n_loci, usually)
          )
{
   
   YY::Stopwatch stopWatch;  double dStartTime = 0.0;  stopWatch.start();   //cout << __FUNCTION__ << ":  stopWatch started." << endl;

   DistSampler distSampler;    // Dispenses random numbers from several Distributions.  (RNG SEED is set in ctor)
   
   const int M                = this->_iSeqLen;        // This should have been set before this .Run function was called.        
   const int n_loci           = this->_iNum_loci;      // This should have been set before this .Run function was called.
	const int iLastBurnInIndex = this->_iBurnInIterations - 2;     // Example:  200-2 == [198]  (which is the 199th position) 

   if(!_curr_draw)                         // If our member variable hasn't been created yet...
   {
      this->CreateCurrDraw( n_loci,M );    // Create our 2D array member. This should only happen the first time we are called.
   }                                       // This will be deleted in our dtor.


   vector<double> pw(n_loci);    
   vector<double> ct(n_loci);

   for(int it=1; it< _iTotalIterations; it++)
   { 
      double ss=0.0; 
      for (int j=0; j<n_loci; j++)    //for(int k=0; k<M; k++) // switch the order of k, j loops
      {
         pw[j] = 0.0;
         ct[j] = 0.1;
         ct[j] += t_v[j];	
	  
         bool* pZj = Z[j];  // for(int uu = 0; uu<36; ++uu)  { std::cout << pZj[uu];  } std::cout << endl;

         double qq;
         bool itsOne = false;
         double dPW = 0.0;
	    
         for (int k=0; k< M; k++) 
         {
            _curr_draw[j][k] = hbetas[it-1][j*M+k];

            if( train[j][k].size() > 0 )
            { 
               size_t tmp_l = train[j][k].size();    // Number of UniqueReads overlaying our MultiRead at position 'k'
               vector<double> u (tmp_l, 0.0);	  		

               vector<int> tmp_mm;
               
               for(int ii=0; ii<(int)tmp_l; ii++)
               {                                          
                  if( train[j][k][ii].boMisMatch == true )     
                     tmp_mm.push_back(ii);                      
               }

               size_t sumz = tmp_mm.size();   // Num of UniqueReads overlaying our MultiRead at position 'k', that have a MisMatch at 'k'. 

               vector<double> q(sumz);    
               vector<double> rr(sumz);   

               for( int ii=0 ;ii<(int)sumz; ii++ )
               {
				      char s = train[j][k][tmp_mm[ii]].cMis;         // The error Profile char
                  double score = a[s];                           // TODO: We need to check to see if 's' is outside the bounds of array 'a'.

                  q[ii]  = (1.0-score) * _curr_draw[j][k] + score;
                  rr[ii] = distSampler.runif( );   
                  if(rr[ii] < (1.0-score)*_curr_draw[j][k] / q[ii])
                  {
                     u[tmp_mm[ii]] = 1 ;
                  }
               }

               double sum_u = 0.0;
               for(int ii=0; ii<(int)tmp_l; ii++)
               {
                  sum_u += u[ii];
               }
               _curr_draw[j][k] = distSampler.rbeta( sum_u + _dPolymorphismRate, double(tmp_l-sumz+1.0) ); 	 
            
            } // endif

            if(it > iLastBurnInIndex) // If we're past the burn-in section...
            {
               char s = xu[j][k];       // Char within the Profile string.
               double score = a[s];     // Lookup in 'a' <map>.
               qq = _curr_draw[j][k] * (1-score) + score; 
               itsOne = pZj[k];
               dPW += ( itsOne ? log(qq) : log(1-qq) );
            }

         } // k loop // previous j loop: 0- (n_loci-1)
	
         if(it > iLastBurnInIndex)     // If we're past the burn-in section...
         {
	         pw[j] = exp(dPW);          //pw[j] = exp(pw[j]);   // previous code
	         ss   += pw[j] * ct[j];
         }
                                                                                   
      } // j loop // previous k loop: 0- (M-1)
  

      for(int ii=0; ii<n_loci*M; ii++)
      {
         hbetas[it][ii] = _curr_draw[ii/M][ii%M];
      }	

      if(it > iLastBurnInIndex) // If we're past the burn-in section... 
      {
         for(int j=0; j<n_loci; j++) 
         {
            prob[j] += pw[j] * ct[j]/ss;
         }
      }
      
   } // it loop
  
    
   for(int j=0; j<n_loci; j++) 
   {
      prob[j] = prob[j] / (_iTotalIterations - _iBurnInIterations  +1);   
      assert(prob[j] >= 0.0 && prob[j] <= 1.0);      // This line is removed by the Microsoft compiler in Release builds.
   }
	
   return  (stopWatch.read() - dStartTime); 

}

