#include <cctype>
#include <algorithm>
#include "StringTools.h"

//namespace MDACC_Biostat
//{
    // --------------------------------------------------------------------------------
    /*! Search through 'sStr', looking for the first occurance of 'sDelim'. \n
        If 'sDelim' is found, 'sResult' will be returned containing the characters
        beginning just after (to the right of) 'sDelim' through the end of 'sStr'.
        If 'sDelim' is NOT found, 'sResult' will contain exactly the characters in 'sStr'.

        @return String containing chars beginning at the start of sStr, and ending just \n
        before the 'sDelim'. If 'sDelim' is NOT found, the return value will be empty.
    */
    std::string
    GetNextToken( std::string const & sStr,      //!< [in]  Input string (searched)
                  std::string const & sDelim,    //!< [in]  Delimiter to look for.
                  std::string       & sResult    //!< [out] String consisting of chars from the end of the delimiter to the end of the string.
                )
    {
        std::string sRtn;
        std::string::size_type iPos;
        iPos = sStr.find( sDelim );
        if( iPos == std::string::npos )      // If the delimiter is not in sStr..
        {
            sResult = sStr;
            //  sRtn is left blank
        }
        else                                 // We found the delimiter
        {
            sRtn    = sStr.substr( 0, iPos );
            sResult = sStr.substr( iPos + sDelim.size() );
        }
        return sRtn;
    }

    // --------------------------------------------------------------------------------
    /*! Split a string into a list of sub-strings.  The sub-strings are expected to
        be separated by 'sDelim'. Each of the parsed sub-strings is placed into
        'lList' and returned.  Note: 'lList' is NOT emptied before the parsing begins.
        @return The quantity of items in 'lList'.
    */
    std::list<std::string>::size_type
    Split( std::string const      & sStrToSplit, //!< [in]  String to split
           std::string const      & sDelim,      //!< [in]  Delimiter that separates items in 'sInput'
           std::list<std::string> & lList        //!< [out] Resulting list of items, after split.
         )
    {
        std::string sInput(sStrToSplit);
        std::string sResult;
        bool boKeepParsing = true;
        while( boKeepParsing )
        {
            std::string sToken = GetNextToken( sInput, sDelim, sResult );

            if(sToken.empty()) // If no token was returned....
            {                  // (the delimiter might be right at the start of the string)
                if( sInput.size() == sResult.size() )   // If no delimiter was found..
                {                                       // Assume the value is there.
                    if( ! sResult.empty() )             // If we've got something to push...
                        lList.push_back( sResult );
                    boKeepParsing = false;              // We've reached the end of the string
                }
                else    // The delimiter was at the FIRST of the string...
                {       //  ... eat the delimiter, then continue parsing.
                    sInput = sResult;
                    continue;
                }
            }
            else    // A Token WAS returned....
            {
                lList.push_back( sToken );
                sInput = sResult;
            }
        }

        return lList.size();
    }

    // --------------------------------------------------------------------------------
    /*! Split a string into a list of sub-strings.  The sub-strings are expected to
        be separated by 'sDelim'. Each of the parsed sub-strings is placed into
        'vItems' and returned.  Note: 'vItems' is NOT emptied before the parsing begins.
        @return The quantity of items in 'vItems'.
    */
    std::vector<std::string>::size_type
    Split( std::string const      & sStrToSplit, //!< [in]  String to split
           std::string const      & sDelim,      //!< [in]  Delimiter that separates items in 'sInput'
           std::vector<std::string> & vItems        //!< [out] Resulting list of items, after split.
         )
    {
        std::string sInput(sStrToSplit);
        std::string sResult;
        bool boKeepParsing = true;
        while( boKeepParsing )
        {
            std::string sToken = GetNextToken( sInput, sDelim, sResult );

            if(sToken.empty()) // If no token was returned....
            {                  // (the delimiter might be right at the start of the string)
                if( sInput.size() == sResult.size() )   // If no delimiter was found..
                {                                       // Assume the value is there.
                    if( ! sResult.empty() )             // If we've got something to push...
                        vItems.push_back( sResult );
                    boKeepParsing = false;              // We've reached the end of the string
                }
                else    // The delimiter was at the FIRST of the string...
                {       //  ... eat the delimiter, then continue parsing.
                    sInput = sResult;
                    continue;
                }
            }
            else    // A Token WAS returned....
            {
                vItems.push_back( sToken );
                sInput = sResult;
            }
        }

        return vItems.size();
    }


    // --------------------------------------------------------------------------------
    /*! Determine whether the last character(s) in the input string (sTargetArg),
        match the entire sSuffix string. 
        If boCaseInsensitive is true, then the search for the sSuffix at the end
        of sTargetArg is performed withOUT regard to case.
        @return true if the sTargetArg ends with sSuffix, false otherwise.
    */
    bool StringEndsWith( std::string const & sTargetArg,   //!< [in]  String to examine
                         std::string sSuffix,              //!< [in]  Does sTargetArg 'end' with this? (NOT a reference!)
                         bool const boCaseInsensitive )    //!< [in]  Disregard case if true (.htm == .HTM)
    {
        bool boFound = false;
        std::string sTarget = sTargetArg;
           
        if( boCaseInsensitive )
        {
            std::transform( sTarget.begin(), sTarget.end(), sTarget.begin(), tolower );
            std::transform( sSuffix.begin(), sSuffix.end(), sSuffix.begin(), tolower );
        }
        std::string::size_type iPos = sTarget.rfind(sSuffix);
        boFound = (iPos != std::string::npos) && (iPos+sSuffix.size() == sTarget.size());
           
        return boFound;
    }    

    //-----------------------------------------------------------------------------
    /*! Change the extension (characters following last dot) of 'sFileName' to
        'sExt'.  The added extension will always contain a dot as the first character,
        even if 'sExt' does not begin with a dot.
        @return This function does not return a value.
    */
    void SetFileExtension( std::string       & sFileName,    //!< [in,out] Filename to change
                           std::string const & sExt )        //!< [in] Extension     
    {
        std::string::size_type iPosFwdSlash  = sFileName.rfind("/");
        std::string::size_type iPosBackSlash = sFileName.rfind("\\");
        std::string::size_type iPosDot       = sFileName.rfind(".");

        if( std::string::npos != iPosDot ) // if we found a dot...
        {
            if(   std::string::npos != iPosFwdSlash    // if we found a slash
               || std::string::npos != iPosBackSlash ) 
            {
                std::string::size_type iPosSlash = 0;
                if(   std::string::npos != iPosFwdSlash )   // if we found a forward slash
                {
                    iPosSlash = ( iPosFwdSlash > iPosSlash ? iPosFwdSlash  : iPosSlash );
                } 
                if(   std::string::npos != iPosBackSlash )   // if we found a back slash
                {
                    iPosSlash = ( iPosBackSlash > iPosSlash ? iPosBackSlash  : iPosSlash );
                } 
                // iPosSlash is the position of the LAST slash, whether forward or backslash.

                if( iPosSlash < iPosDot )  // if the dot is to the RIGHT of the slash..
                {
                    sFileName.resize(iPosDot);   // chop off the old dot+extension 
                }
                //else the dot is way to the left within the path (it is not a separator between the filename and the extension)
            }
            else   // no slash found
            {
                sFileName.resize(iPosDot);   // chop off the old dot+extension 
            }
        }
        // The filename should now NOT have a trialing dot.

        // Add the new extension:
        if(sExt.at(0) == '.')
            sFileName += sExt;
        else
            sFileName += "." + sExt;
    }    


//}  // namespace MDACC_Biostat
