
#include <algorithm>
#include <cassert>
#include <climits>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>
#include <map>
#include <omp.h>

using namespace std;

#include "constants.h"
#include "Read.h"
#include "TrainData.h"

#include "distsampling.h"

#include "utils.h"
#include "MC2.h"
#include "YY_stopwatch.h"

#include "BMTypes.h"
#include "BmCalc.h"
#include "ThreadDataMgr.h"
#include "Logger.h"


// ----------------------------------------------------------------------------
// This function prepares input data for _iNumThread number of 'jobs', each of
//   which we'll run on a separate thread of execution.
// The job input information is placed into BmCalc objects, and each of the
//  BmCalc objects is told to ".Run".
// The calculation results are written to files whose names are specified in the
//  ThreadDataMgr::PopulateJobInfoVector  function.
// 
// map_1 is the container of unique Reads.
// mapMultiRead is the container of MultiReads, each entry containing 2 to 5 Reads having the same ID.
//
void DataTrainer::trainData( 
                MultiReadVector    & map_1,               // Unique Reads
                MultiReadVector    & mapMultiRead,        // MultiReads
                ErrorScoreMap      & a,                   // map<char,double>
                std::vector<int>   & positionV,            // Positions of each of the UniqueReads (sorted)
				    string sProbOutput,
				    bool boStrandness //whether consider strandness when mapping uniqueReads to multiRead
                )
{
   if(mapMultiRead.size() < 1)
      return;    // Nothing to do.

   Logger & lg = GetLogger();        // Fetch our global Logger object.  (Declaration MUST be a reference!  Logger is not assignable.)
   std::ostringstream os;

   int iNumParts = _iNParts;
   int iPart     = _iPart;    // These are the *local* variable we'll use.
                              //  They may be different from _iNParts and _iPart, if the mapSize is too small.

   YY::Stopwatch stopWatch;  double dTotalTime = 0.0;  stopWatch.start();   //cout << __FUNCTION__ << ":  stopWatch started." << endl;

   int n_loci = mapMultiRead[0]->Size();
   //lg.Write( ((os << endl << __FUNCTION__ << ": The size of mapMultiRead " << n_loci << " is: (" << mapMultiRead.size() << ") entries." << endl), os));

   // We are running the calculations on one '_iPart' of several '_iNParts'.  
   //  ( a _iPart means a collection of consecutive MultiReads).
   //  We first determine which iPartStartIndex and iPartEndIndex indices within the mapMultiRead collection we will
   //   perform the calculations over.

   // _iPart is ONE-based !!  
   int iMapSize;
   int iNumItemsPerPart, iPartStartIndex, iPartEndIndex;
   
   iMapSize  = (int)mapMultiRead.size();    

   if(iMapSize <= _iNParts)  // NOTE:
   {                         // It the mapSize is too small, we can't divide-up the parts effectively.
      iNumParts = 1;         // So we just alter the iNumParts and iPart
      iPart     = 1; 
   }
   
   iNumItemsPerPart  = iMapSize / iNumParts;            // Integer, with possible leftovers.
   iPartStartIndex   = iNumItemsPerPart * (iPart-1);  // 
   iPartEndIndex     = iPartStartIndex  + iNumItemsPerPart ;   // Our iterations will run (i=iStartIndex; < iEndIndex; ++i)

   if(iPart == iNumParts)    // If we are doing the last part... pick up any leftovers due to iMapSize/iNumParts integer division.
   {
      //cout << __FUNCTION__ << ": We're on the last part, iPart(" << iPart << ") == iNumParts(" << iNumParts << ")." << endl;
      //cout << __FUNCTION__ << ": iPartEndIndex is (" << iPartEndIndex << "),  iMapSize is (" << iMapSize << ")" << endl;
      if(iPartEndIndex < iMapSize)  
      { 
         //cout << __FUNCTION__ << ": Adjusting iPartEndIndex to match iMapSize."  << endl;
         iPartEndIndex = iMapSize;
      }
   }

   int iNumJobs = _iNumThreads;     // This is the number of software Threads we'll start.
   int iNumItemsToCalculate =iPartEndIndex  - iPartStartIndex; 
   if(iNumJobs > iNumItemsToCalculate )  iNumJobs = iNumItemsToCalculate;    // NOTE: this !

   lg.Write( ((os << __FUNCTION__ << ": We'll run with (" << iNumJobs << ") threads." << endl), os) );
   
   ThreadDataMgr thrdMgr;          // Keeps track of per-thread info for our BmCalculator object.

   thrdMgr.PopulateJobInfoVector(iPartStartIndex, iPartEndIndex, iNumJobs, n_loci, _iPart );

   size_t iNJ = thrdMgr._vJobInfos.size();

   // Build a vector of calculators, then run each on a separate thread:
   std::vector<BmCalc> vBmCalcs;
   for(int j=0; j<iNumJobs; ++j)
   {
      BmCalc bmCalculator( _iSeqLen, _iSeqFlankingValue, _dPolymorphismRate, _iTotalIterations, _iBurninIterations); // Set properties via the ctor.
      vBmCalcs.push_back(bmCalculator);     
   }

   // Use OpenMP here, to run each calculation on a separate thread.
   int j=0;  
   //omp_set_num_threads(iNumJobs);
   #pragma omp parallel for private (j)
   for( j=0; j<iNumJobs; ++j)
   {
      vBmCalcs[j].Run( map_1,  mapMultiRead, a, positionV, *(thrdMgr._vJobInfos[j]), sProbOutput, boStrandness );     
   }                        

   if(iNumJobs>0)
   {
      thrdMgr.MergeOutputFiles();
   }

   // I'm counting on the dtor for the thrdMgr to delete the jobs in thrdMgr._vJobInfos

   dTotalTime = stopWatch.read();   
   
   lg.Write( ((os << __FUNCTION__ << ":  Total elapsed time in this function: " << dTotalTime << " secs, (" << dTotalTime/60.0 << " mins)" << endl), os));
}

