//
// DISTSAMPLING.CPP
//

#include <algorithm>
#include <climits>
#include <cmath>
#include <ctime>
#include <iostream>
#include <iterator>
#include <numeric>
#include <vector>
using namespace std;
#include "distsampling.h"


// ----------------------------------------------------------------------------
// Return a random number such that (0.0 > num > 1.0), from a Uniform Distribution.  
double DistSampler::runif( )                      //double startv, double endv, SEED_TYPE* idum )
{
   // Note: we are ignoring all the args.

   //if(startv > 0.0  || endv > 1.0)  // We don't use these args, but I'd like to know if someone is trying send them to us.
   //   throw std::exception(__FUNCTION__": Requests are supported for Uniform random numbers between (0.0, 1.0) only.");

   double dUnif = this->_rng.GetUniform();
   return dUnif;

#if(0)
   const int IA = 16807;
   const int IM = 2147483647;  // INT_MAX
   const double AM = (1.0/IM);
   const int IQ = 127773;
   const int IR = 2836;
   const int NTAB = 32;
   const int NDIV = (1+(IM-1)/NTAB);
   const double EPS = 1.2e-7;
   const double RNMX = (1.0-EPS);

   int j;
   long k;
   static SEED_TYPE iy = 0;
   static SEED_TYPE iv[NTAB];


   if (*idum <= 0 || !iy) 
   {
      if (-(*idum) < 1) 
      {
         *idum = 1;
      }
      else 
      {
         *idum = -(*idum);
      }

      for (j = NTAB+7; j >= 0; j--) 
      {
         k = (*idum) / IQ;
         *idum = IA * (*idum - k*IQ) - IR*k;
         if (*idum < 0) 
         {
               *idum += IM;
         }
         if (j < NTAB) 
         {
               iv[j] = *idum;
         }
      }
      iy = iv[0];
   }

   k = (*idum) / IQ;
   *idum = IA * (*idum-k*IQ) - IR*k;
   if (*idum < 0) 
   {
      *idum += IM;
   }
   j = iy / NDIV;
   iy = iv[j];
   iv[j] = *idum;

   double fint = AM * iy;
   if (fint > RNMX) 
   {
      fint = RNMX;
   }

   return fint * (endv-startv) + startv;
#endif
}


// ----------------------------------------------------------------------------
double DistSampler::rgamma(double shape, double rate )   // , SEED_TYPE* idum)
{
   
#define EXP1  2.7182818284590451       // double EXP1 = exp(1.0);
   int shapei = static_cast<int>(floor(shape));
   double shapef = shape - shapei;
   double x = 1.0;

   if (shapei < 6) 
   {
      for (int j = 0; j < shapei; j++)  
      {
         x *= runif( );   // 0.0, 1.0, idum);
      }
      x = -log(x);
   }
   else 
   {
      double e;
      do 
      {
         double am, s, y;
         do 
         {
            double v1, v2;
            do 
            {
               v1 =       runif( );    // 0.0, 1.0, idum);
               v2 = 2.0 * runif( );    // 0.0, 1.0, idum) - 1.0;
            } while (v1*v1+v2*v2 > 1.0);
   
            y = v2 / v1;
            am = shapei-1;
            s = sqrt(2.0*am+1.0);
            x = s*y+am;
         } while (x <= 0.0);
   
         e = (1.0+y*y) * exp(am*log(x/am)-s*y);
      } while (runif() > e);       // (runif(0.0, 1.0, idum) > e);
   }

   //generate the other part
  
   const int size = 3;   
   double V[size];       
   double v0 = EXP1 / (shapef +EXP1) ;
   double eps, ita;

   do 
   {        
      for (int m1 = 0; m1 < size; m1++) 
      {
         V[m1] = runif( );   // 0.0, 1.0, idum);
      }

      if (V[0] <= v0)  
      {
         eps = pow(V[1], 1.0/shapef);
         ita = V[2] * pow(eps, shapef-1);
      }
      else 
      {
         eps = 1 - log(V[1]);
         ita = V[2] * exp(-eps);    // V[2] * exp(-1.0*eps);
      }
   } while  (ita > (pow(eps, shapef-1) * exp(-eps)));     // (ita > (pow(eps, shapef-1) * exp(-1.0*eps)));
 
   return (eps+x) / rate;
}


// ----------------------------------------------------------------------------
double DistSampler::rbeta(double alpha, double beta )   // , SEED_TYPE* idum)
{
    double x1 = rgamma(alpha, 10.0);  // , idum);
    double x2 = rgamma(beta , 10.0);  // , idum);

    return x1 / (x1+x2);
}


#if(0)

// ----------------------------------------------------------------------------
double rnorm(double mu, double sd, SEED_TYPE* idum)
{
    static bool iset = false;
    static double gset;

    if (!iset) {
        double rsq;
        double v1, v2;

        do {
            v1 = 2.0*runif(0, 1, idum)-1.0;
            v2 = 2.0*runif(0, 1, idum)-1.0;
            rsq = v1*v1+v2*v2;
        } while (rsq >= 1.0 || rsq == 0.0);

        double fac = sqrt(-2.0*log(rsq)/rsq);
        gset = v1 * fac;
        iset = true;
        return v2*fac*sd+mu;
    }
    else {
        iset = false;
        return gset*sd+mu;
    }
}



// ----------------------------------------------------------------------------
int sample(vector<double>& p, SEED_TYPE* idum)
{
    double sum = 0.0;
    size_t n = p.size();
    
    for (int i = 0; i < (int)n; i++) {
        if (p[i] < 0)
            cerr << "sample: probabilities should be positive" << endl;
        else
            sum += p[i];
    }
    for (int i = 0; i < (int)n; i++) {
        p[i] = p[i] / sum;
    }

    int fin = 0;
    if (n != 1) {
        double u = runif(0.0, 1.0, idum);
        double accum = 0.0;
        for (int i = 0; i < (int)n; i++) {
            accum += p[i];
            if (accum >= u) {
                fin = i;
                break;
            }
        }
    }

    return fin;
}


// ----------------------------------------------------------------------------
void rdirichlet(vector<int>& nk, vector<double>& d, SEED_TYPE* idum)
{
    size_t K = nk.size();
    vector<double> x(K, 0.0);

    for (int k = 0; k < (int)K; k++) {
        x[k] = rgamma(nk[k], 1.0, idum);
    }

    double sum=std::accumulate(x.begin(), x.end(), 0.0);

    for (int k = 0; k < (int)K; k++) {
        d[k] = x[k] / sum;
    }
}


/*
    Cholesky decomposition can be done in-place.  Why I did not
    implement it that way:
        - It makes the code a little uglier,
        - Sometimes you want the original matrix to not be destroyed,
        - I'm not interested in matrices seriously larger
        than 100x100, which is 80k, so the wastage of storage
        is a total nonissue, especially in C.
*/

// ----------------------------------------------------------------------------
void cholesky(vector<vector<double> >& a)
{
    size_t n = a.size();
    vector<double> diagonal(n, 0.0);

    for (int i = 0; i < (int)n; i++) {
      for (int j = i; j < (int)n; j++) {
            double sum;

            sum = a[i][j];
            for (int k = i-1; k >= 0; k--) {
                sum -= a[i][k] * a[j][k];
            }
            if (i == j) {
                if (sum <= 0.0)
                    cerr << "cholesky: matrix not positive definite" << endl;
                diagonal[i] = sqrt(sum);
            }
            else {
                a[j][i] = sum / diagonal[i];
                a[i][j] = a[j][i];
            }
        }
    }

    for (int i = 0; i < (int)n; i++) {
        a[i][i] = diagonal[i];
    }
}
#endif  // (0)

